//=============================================================================
//  Unified IO - C++ interface for industrial IO cards
//  Copyright (C) 2000, UAB "BBD SOFT" ( http://www.bbdsoft.com/ )
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
//  The author of this program may be contacted at developers@bbdsoft.com
//=============================================================================


#ifndef _BTRACE_
   #include "btrace.hpp"
#endif

#ifndef _PARALLELPORT_
   #include "parallelport.hpp"
#endif

#ifndef _STDEXCEPT_
   #include <stdexcept>
#endif

#ifndef _CSTRING_
   #include <cstring>
#endif

#ifndef _IOPORT_
   #include "ioport.hpp"
#endif

#ifndef _IOTRACK_
   #include "iotrack.hpp"
#endif

#ifndef _PHYSMEM_
   #include "physmem.hpp"
#endif

using namespace std;
using namespace Uniio;

//-----------------------------------------------------------------------------
ParallelPort::ParallelPort ( const char * inName )
  : IOCard ("LPT")
  , theBaseIOPort ( 0 )
{
BFUNCTRACE_DEVELOP ();
   PhysicalMemory aMemory (0x00400, 0x11C, false);

   if ( !inName || !strcmp(inName, "prn") || !strcmp(inName, "lpt1")
                || !strcmp(inName, "PRN") || !strcmp(inName, "LPT1")
      )
   {
      theBaseIOPort = aMemory.readShort (0x08);
   }
   else
   if (  !strcmp(inName, "lpt2")
      || !strcmp(inName, "LPT2")
      )
   {
      theBaseIOPort = aMemory.readShort (0x0A);
   }
   else
   if (  !strcmp(inName, "lpt3")
      || !strcmp(inName, "LPT3")
      )
   {
      theBaseIOPort = aMemory.readShort (0x0C);
   }
   else
   if (  !strcmp(inName, "lpt4")
      || !strcmp(inName, "LPT4")
      )
   {
      theBaseIOPort = aMemory.readShort (0x0E);
   }
   if (!theBaseIOPort)
   {
      logic_error exc ("ParallelPort: specified parallel port does not exists.");
      throw exc;
   } // endif
     
}  // end constructor

//-----------------------------------------------------------------------------
ParallelPort::ParallelPort ( const unsigned short inBaseIOPort )
  : IOCard ("LPT")
  , theBaseIOPort ( inBaseIOPort )
{
BFUNCTRACE_DEVELOP ();
   PhysicalMemory aMemory (0x00400, 0x11C, false);

if (   theBaseIOPort == 0
   || (theBaseIOPort != aMemory.readShort (0x08) 
    && theBaseIOPort != aMemory.readShort (0x0A)
    && theBaseIOPort != aMemory.readShort (0x0C)
    && theBaseIOPort != aMemory.readShort (0x0E)
      )
   )
{
   logic_error exc ("ParallelPort: invalid base IO port number specified or port does not exist.");
   throw exc;
} // endif 
     
}  // end constructor


//-----------------------------------------------------------------------------
ParallelPort::~ParallelPort ()
{
BFUNCTRACE_DEVELOP ();
} // end destructor


//-----------------------------------------------------------------------------
IOCard & ParallelPort::updateInputs ( const IOTrack & inTrack
                               , long * inImage
                               , long & inTrackState
                               )
{
BFUNCTRACE_DEVELOP ();

   // to increase processing speed update only banks that have at least one input
   if ( inTrack.hasPointOfType(IOTrack::digitalInput) )
   {

      unsigned char newValue = 0;
      // first 8 databits are outputs

      int j = 8;
      if (inTrack.pointType(j) == IOTrack::digitalInput)
      {
         // we have 4 control inputs
         IOPort dataRegister ( baseIOPort()+controlOffset
                             , false               // write access not required
                             );
         newValue = dataRegister.readChar ();
         // invert one bit
         if (newValue & 0x4)
         {
            newValue &= ~0x4;
         }
         else
         {
            newValue |= 0x4;
         }
      }
      else
      {
         // skip 4 outputs
         j += 4;
      }
      for (; j<inTrack.numberOfPoints(); j++)
      {
         if ( j==12 )
         {
         // we have 5 status inputs
         IOPort dataRegister ( baseIOPort()+statusOffset
                             , false               // write access not required
                             );
         newValue = dataRegister.readChar ();
         // invert one bit
         if (newValue & 0x80)
         {
            newValue &= ~0x80;
         }
         else
         {
            newValue |= 0x80;
         }
         newValue = ~(newValue >> 3);
         }
         if (inTrack.pointType(j) == IOTrack::digitalInput)
         {
            inImage[j] = newValue %2;
         } // end if
         newValue /= 2;
      } // endfor
   } // endif

return * this;
} // end ParallelPort::updateInputs (..)


//-----------------------------------------------------------------------------
IOCard & ParallelPort::updateOutputs ( const IOTrack & inTrack
                                , const long * const inImage
                                , long & inTrackState
                                )
{
BFUNCTRACE_DEVELOP ();

   // to increase processing speed update only banks that have at least one input
   if ( inTrack.hasPointOfType(IOTrack::digitalOutput) )
   {
      unsigned char newValue = 0;
      int j = 0;
      for ( 
          ; j<12
          ; j++)
      {
         if ( inImage[j] && (inTrack.pointType(j) == IOTrack::digitalOutput))
         {
            newValue |= (1 << (j%8));
         } // endif
         if ( j == 7 )
         {
            // first 8 databits
            IOPort dataRegister ( baseIOPort() );
            dataRegister.writeChar (~newValue);
            newValue = 0;

            if (inTrack.pointType(j+1) == IOTrack::digitalInput)
            {
               break;
            }
         } // end if
      } // endfor

      if ( j > 7 )
      {
         // we have 4 control outputs
         // invert one bit
         if (newValue & 0x4)
         {
            newValue &= ~0x4;
         }
         else
         {
            newValue |= 0x4;
         }
         IOPort dataRegister ( baseIOPort()+controlOffset );
         dataRegister.writeChar (newValue);
      } // end if
   } // endif

return * this;
} // end ParallelPort::updateOutputs (..)


//-----------------------------------------------------------------------------
IOCard & ParallelPort::addTrack ( IOTrack & inTrack )
{
BFUNCTRACE_DEVELOP ();


if (   inTrack.hasPointOfType ('U')
    || inTrack.hasPointOfType ('A') 
    || inTrack.hasPointOfType ('C') 
    || inTrack.hasPointOfType ('R') 
   )
{
   throw logic_error ("ParallelPort:  does not support analog I/O and counters.");
} // endif

     if (inTrack.numberOfPoints() != maxNumberOfPoints)
      {
         logic_error exc ("ParallelPort track can be only 17 points long.");
         throw exc;
      } // endif

for ( int i = 0
    ; i < inTrack.numberOfPoints()
    ; i++
    )
{
    if ( (i < 8) && (inTrack.pointType (i) == IOTrack::digitalInput) )
    {
         logic_error exc ("ParallelPort track first 8 points can not be Inputs.");
         throw exc;
    }
    if ( (i < 12) && ( i > 8 ) && (inTrack.pointType(i) != inTrack.pointType(i-1)) )
    {
         logic_error exc ("ParallelPort track points 8-12 can be only Inputs or outputs at the same time.");
         throw exc;
    }
    if ( (i >= 12 ) && (inTrack.pointType (i) == IOTrack::digitalOutput) )
    {
         logic_error exc ("ParallelPort track points 13-17 can not be Outputs.");
         throw exc;
    }
} // end for


   if (inTrack.pointType(8) == IOTrack::digitalInput)
   {
      IOPort dataRegister (baseIOPort() + controlOffset );
      char newValue = dataRegister.readChar();
      for (int i =0; i< 4; i++)
      {
         if (inTrack.pointType(i+8) == IOTrack::digitalInput)
         {
            if (i == 2)
               newValue |= (1 << i);
            else
               newValue &= ~(1 << i);
         }
      }
      dataRegister.writeChar (newValue);
   }

return * this;
} // end ParallelPort::addTrack (..)
